-module(howdy@authentication@cookie).
-compile(no_auto_import).

-export([new/0, new_with_config/1, authenticate_with_cookie/1, sign_in/5, sign_out/2]).
-export_type([session_func_ets/0, session_storage/0, timeouts/0, cookie_config/0]).

-type session_func_ets() :: session_func_ets.

-type session_storage() :: {session_storage,
                            fun((binary(), binary(), integer(), list({binary(),
                                                                      binary()})) -> boolean()),
                            fun((binary()) -> {ok,
                                               {binary(),
                                                binary(),
                                                integer(),
                                                list({binary(), binary()})}} |
                                {error, nil}),
                            fun((binary()) -> boolean())}.

-type timeouts() :: {timeouts, integer(), integer()}.

-type cookie_config() :: {cookie_config,
                          gleam@option:option(gleam@http@cookie:attributes()),
                          gleam@option:option(session_storage()),
                          gleam@option:option(timeouts())}.

-spec new() -> nil.
new() ->
    new_with_config({cookie_config, none, none, none}).

-spec new_with_config(cookie_config()) -> nil.
new_with_config(Cookie_config) ->
    howdy@authentication@ets:new(session_func_ets, [set, public, named_table]),
    case erlang:element(2, Cookie_config) of
        {some, Cookie} ->
            howdy@authentication@ets:insert(
                session_func_ets,
                {<<"cookie"/utf8>>, Cookie}
            ),
            nil;

        none ->
            nil
    end,
    case erlang:element(3, Cookie_config) of
        {some, Storage} ->
            howdy@authentication@ets:insert(
                session_func_ets,
                {<<"insert"/utf8>>, erlang:element(2, Storage)}
            ),
            howdy@authentication@ets:insert(
                session_func_ets,
                {<<"lookup"/utf8>>, erlang:element(3, Storage)}
            ),
            howdy@authentication@ets:insert(
                session_func_ets,
                {<<"delete"/utf8>>, erlang:element(4, Storage)}
            ),
            nil;

        none ->
            howdy@authentication@session:new(),
            nil
    end.

-spec authenticate_with_cookie(howdy@context:context(any())) -> {ok,
                                                                 howdy@context@user:user()} |
    {error, nil}.
authenticate_with_cookie(Context) ->
    _pipe = Context,
    _pipe@1 = get_cookie(_pipe),
    find_user(_pipe@1).

-spec sign_in(
    binary(),
    binary(),
    boolean(),
    gleam@http@response:response(JBT),
    fun((binary(), binary()) -> {ok, list({binary(), binary()})} |
        {error, any()})
) -> {ok, gleam@http@response:response(JBT)} | {error, nil}.
sign_in(Username, Password, Remember_me, Resp, Fun) ->
    {_@1, Now, _@2} = erlang:timestamp(),
    Timeout = get_timeout(Remember_me) + Now,
    case Fun(Username, Password) of
        {ok, Claims} ->
            Id = get_unique_key(),
            insert(Id, Username, Timeout, Claims),
            _pipe = Resp,
            _pipe@1 = gleam@http@response:set_cookie(
                _pipe,
                <<"howdy:user_session"/utf8>>,
                Id,
                get_cookie_attributes()
            ),
            {ok, _pipe@1};

        {error, _@3} ->
            {error, nil}
    end.

-spec sign_out(howdy@context:context(any()), gleam@http@response:response(JCE)) -> gleam@http@response:response(JCE).
sign_out(Context, Resp) ->
    case get_cookie(Context) of
        {ok, Cookie} ->
            delete(erlang:element(2, Cookie));

        {error, _@1} ->
            false
    end,
    _pipe = Resp,
    gleam@http@response:expire_cookie(
        _pipe,
        <<"howdy:user_session"/utf8>>,
        get_cookie_attributes()
    ).

-spec get_cookie(howdy@context:context(any())) -> {ok, {binary(), binary()}} |
    {error, nil}.
get_cookie(Context) ->
    _pipe = gleam@http@request:get_cookies(erlang:element(3, Context)),
    gleam@list:find(
        _pipe,
        fun(Cookie) ->
            erlang:element(1, Cookie)
            =:= <<"howdy:user_session"/utf8>>
        end
    ).

-spec find_user({ok, {binary(), binary()}} | {error, any()}) -> {ok,
                                                                 howdy@context@user:user()} |
    {error, nil}.
find_user(Cookie_result) ->
    case Cookie_result of
        {ok, Cookie} ->
            get_user(erlang:element(2, Cookie));

        {error, _@1} ->
            {error, nil}
    end.

-spec get_user(binary()) -> {ok, howdy@context@user:user()} | {error, nil}.
get_user(Data) ->
    case lookup(Data) of
        {ok, Result} ->
            {_@1, Name, Timeout, Claims} = Result,
            {_@2, Now, _@3} = erlang:timestamp(),
            case Timeout > Now of
                true ->
                    {ok, {user, Name, Claims}};

                false ->
                    delete(Data),
                    {error, nil}
            end;

        {error, _@4} ->
            {error, nil}
    end.

-spec get_timeout(boolean()) -> integer().
get_timeout(Remember_me) ->
    {_@1, Timeouts} = begin
        _pipe = howdy@authentication@ets:lookup(
            session_func_ets,
            <<"timeouts"/utf8>>
        ),
        _pipe@1 = gleam@list:first(_pipe),
        gleam@result:unwrap(
            _pipe@1,
            {<<"timeouts"/utf8>>, {timeouts, 2592000, 14400}}
        )
    end,
    case Remember_me of
        true ->
            erlang:element(3, Timeouts);

        false ->
            erlang:element(2, Timeouts)
    end.

-spec get_cookie_attributes() -> gleam@http@cookie:attributes().
get_cookie_attributes() ->
    {_@1, Cookie_attributes} = begin
        _pipe = howdy@authentication@ets:lookup(
            session_func_ets,
            <<"cookie"/utf8>>
        ),
        _pipe@1 = gleam@list:first(_pipe),
        gleam@result:unwrap(
            _pipe@1,
            {<<"cookie"/utf8>>, gleam@http@cookie:defaults(http)}
        )
    end,
    Cookie_attributes.

-spec get_unique_key() -> binary().
get_unique_key() ->
    gleam@string:replace(howdy@uuid:v4_string(), <<"-"/utf8>>, <<""/utf8>>).

-spec insert(binary(), binary(), integer(), list({binary(), binary()})) -> boolean().
insert(Key, Username, Expiry, Claims) ->
    {_@1, Inner_insert} = begin
        _pipe = howdy@authentication@ets:lookup(
            session_func_ets,
            <<"insert"/utf8>>
        ),
        _pipe@1 = gleam@list:first(_pipe),
        gleam@result:unwrap(
            _pipe@1,
            {<<"insert"/utf8>>, fun howdy@authentication@session:insert/4}
        )
    end,
    Inner_insert(Key, Username, Expiry, Claims).

-spec lookup(binary()) -> {ok,
                           {binary(),
                            binary(),
                            integer(),
                            list({binary(), binary()})}} |
    {error, nil}.
lookup(Key) ->
    {_@1, Inner_lookup} = begin
        _pipe = howdy@authentication@ets:lookup(
            session_func_ets,
            <<"lookup"/utf8>>
        ),
        _pipe@1 = gleam@list:first(_pipe),
        gleam@result:unwrap(
            _pipe@1,
            {<<"lookup"/utf8>>, fun howdy@authentication@session:lookup/1}
        )
    end,
    Inner_lookup(Key).

-spec delete(binary()) -> boolean().
delete(Key) ->
    {_@1, Inner_delete} = begin
        _pipe = howdy@authentication@ets:lookup(
            session_func_ets,
            <<"delete"/utf8>>
        ),
        _pipe@1 = gleam@list:first(_pipe),
        gleam@result:unwrap(
            _pipe@1,
            {<<"delete"/utf8>>, fun howdy@authentication@session:delete/1}
        )
    end,
    Inner_delete(Key).
